@@ -16,7 +16,11 @@
 
 package io.netty.buffer;
 
+import static io.netty.util.internal.ObjectUtil.checkPositiveOrZero;
+
+import io.netty.util.ByteProcessor;
 import io.netty.util.internal.EmptyArrays;
+import io.netty.util.internal.ObjectUtil;
 import io.netty.util.internal.PlatformDependent;
 import io.netty.util.internal.StringUtil;
 
@@ -25,6 +29,7 @@ import java.io.OutputStream;
 import java.nio.ByteBuffer;
 import java.nio.ByteOrder;
 import java.nio.ReadOnlyBufferException;
+import java.nio.channels.FileChannel;
 import java.nio.channels.GatheringByteChannel;
 import java.nio.channels.ScatteringByteChannel;
 import java.nio.charset.Charset;
@@ -34,6 +39,7 @@ import java.nio.charset.Charset;
  */
 public final class EmptyByteBuf extends ByteBuf {
 
+    static final int EMPTY_BYTE_BUF_HASH_CODE = 1;
     private static final ByteBuffer EMPTY_BYTE_BUFFER = ByteBuffer.allocateDirect(0);
     private static final long EMPTY_BYTE_BUFFER_ADDRESS;
 
@@ -59,11 +65,7 @@ public final class EmptyByteBuf extends ByteBuf {
     }
 
     private EmptyByteBuf(ByteBufAllocator alloc, ByteOrder order) {
-        if (alloc == null) {
-            throw new NullPointerException("alloc");
-        }
-
-        this.alloc = alloc;
+        this.alloc = ObjectUtil.checkNotNull(alloc, "alloc");
         this.order = order;
         str = StringUtil.simpleClassName(this) + (order == ByteOrder.BIG_ENDIAN? "BE" : "LE");
     }
@@ -93,6 +95,16 @@ public final class EmptyByteBuf extends ByteBuf {
         return null;
     }
 
+    @Override
+    public ByteBuf asReadOnly() {
+        return Unpooled.unmodifiableBuffer(this);
+    }
+
+    @Override
+    public boolean isReadOnly() {
+        return false;
+    }
+
     @Override
     public boolean isDirect() {
         return true;
@@ -105,10 +117,7 @@ public final class EmptyByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf order(ByteOrder endianness) {
-        if (endianness == null) {
-            throw new NullPointerException("endianness");
-        }
-        if (endianness == order()) {
+        if (ObjectUtil.checkNotNull(endianness, "endianness") == order()) {
             return this;
         }
 
@@ -210,9 +219,7 @@ public final class EmptyByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf ensureWritable(int minWritableBytes) {
-        if (minWritableBytes < 0) {
-            throw new IllegalArgumentException("minWritableBytes: " + minWritableBytes + " (expected: >= 0)");
-        }
+        checkPositiveOrZero(minWritableBytes, "minWritableBytes");
         if (minWritableBytes != 0) {
             throw new IndexOutOfBoundsException();
         }
@@ -221,9 +228,7 @@ public final class EmptyByteBuf extends ByteBuf {
 
     @Override
     public int ensureWritable(int minWritableBytes, boolean force) {
-        if (minWritableBytes < 0) {
-            throw new IllegalArgumentException("minWritableBytes: " + minWritableBytes + " (expected: >= 0)");
-        }
+        checkPositiveOrZero(minWritableBytes, "minWritableBytes");
 
         if (minWritableBytes == 0) {
             return 0;
@@ -252,36 +257,71 @@ public final class EmptyByteBuf extends ByteBuf {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public short getShortLE(int index) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public int getUnsignedShort(int index) {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public int getUnsignedShortLE(int index) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public int getMedium(int index) {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public int getMediumLE(int index) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public int getUnsignedMedium(int index) {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public int getUnsignedMediumLE(int index) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public int getInt(int index) {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public int getIntLE(int index) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public long getUnsignedInt(int index) {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public long getUnsignedIntLE(int index) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public long getLong(int index) {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public long getLongLE(int index) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public char getChar(int index) {
         throw new IndexOutOfBoundsException();
@@ -338,6 +378,18 @@ public final class EmptyByteBuf extends ByteBuf {
         return 0;
     }
 
+    @Override
+    public int getBytes(int index, FileChannel out, long position, int length) {
+        checkIndex(index, length);
+        return 0;
+    }
+
+    @Override
+    public CharSequence getCharSequence(int index, int length, Charset charset) {
+        checkIndex(index, length);
+        return null;
+    }
+
     @Override
     public ByteBuf setBoolean(int index, boolean value) {
         throw new IndexOutOfBoundsException();
@@ -353,21 +405,41 @@ public final class EmptyByteBuf extends ByteBuf {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public ByteBuf setShortLE(int index, int value) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public ByteBuf setMedium(int index, int value) {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public ByteBuf setMediumLE(int index, int value) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public ByteBuf setInt(int index, int value) {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public ByteBuf setIntLE(int index, int value) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public ByteBuf setLong(int index, long value) {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public ByteBuf setLongLE(int index, long value) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public ByteBuf setChar(int index, int value) {
         throw new IndexOutOfBoundsException();
@@ -425,11 +497,22 @@ public final class EmptyByteBuf extends ByteBuf {
         return 0;
     }
 
+    @Override
+    public int setBytes(int index, FileChannel in, long position, int length) {
+        checkIndex(index, length);
+        return 0;
+    }
+
     @Override
     public ByteBuf setZero(int index, int length) {
         return checkIndex(index, length);
     }
 
+    @Override
+    public int setCharSequence(int index, CharSequence sequence, Charset charset) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public boolean readBoolean() {
         throw new IndexOutOfBoundsException();
@@ -450,36 +533,71 @@ public final class EmptyByteBuf extends ByteBuf {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public short readShortLE() {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public int readUnsignedShort() {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public int readUnsignedShortLE() {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public int readMedium() {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public int readMediumLE() {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public int readUnsignedMedium() {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public int readUnsignedMediumLE() {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public int readInt() {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public int readIntLE() {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public long readUnsignedInt() {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public long readUnsignedIntLE() {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public long readLong() {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public long readLongLE() {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public char readChar() {
         throw new IndexOutOfBoundsException();
@@ -505,6 +623,11 @@ public final class EmptyByteBuf extends ByteBuf {
         return checkLength(length);
     }
 
+    @Override
+    public ByteBuf readRetainedSlice(int length) {
+        return checkLength(length);
+    }
+
     @Override
     public ByteBuf readBytes(ByteBuf dst) {
         return checkLength(dst.writableBytes());
@@ -546,6 +669,18 @@ public final class EmptyByteBuf extends ByteBuf {
         return 0;
     }
 
+    @Override
+    public int readBytes(FileChannel out, long position, int length) {
+        checkLength(length);
+        return 0;
+    }
+
+    @Override
+    public CharSequence readCharSequence(int length, Charset charset) {
+        checkLength(length);
+        return StringUtil.EMPTY_STRING;
+    }
+
     @Override
     public ByteBuf skipBytes(int length) {
         return checkLength(length);
@@ -566,21 +701,41 @@ public final class EmptyByteBuf extends ByteBuf {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public ByteBuf writeShortLE(int value) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public ByteBuf writeMedium(int value) {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public ByteBuf writeMediumLE(int value) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public ByteBuf writeInt(int value) {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public ByteBuf writeIntLE(int value) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public ByteBuf writeLong(long value) {
         throw new IndexOutOfBoundsException();
     }
 
+    @Override
+    public ByteBuf writeLongLE(long value) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public ByteBuf writeChar(int value) {
         throw new IndexOutOfBoundsException();
@@ -598,7 +753,7 @@ public final class EmptyByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf writeBytes(ByteBuf src) {
-        throw new IndexOutOfBoundsException();
+        return checkLength(src.readableBytes());
     }
 
     @Override
@@ -638,11 +793,22 @@ public final class EmptyByteBuf extends ByteBuf {
         return 0;
     }
 
+    @Override
+    public int writeBytes(FileChannel in, long position, int length) {
+        checkLength(length);
+        return 0;
+    }
+
     @Override
     public ByteBuf writeZero(int length) {
         return checkLength(length);
     }
 
+    @Override
+    public int writeCharSequence(CharSequence sequence, Charset charset) {
+        throw new IndexOutOfBoundsException();
+    }
+
     @Override
     public int indexOf(int fromIndex, int toIndex, byte value) {
         checkIndex(fromIndex);
@@ -668,23 +834,23 @@ public final class EmptyByteBuf extends ByteBuf {
     }
 
     @Override
-    public int forEachByte(ByteBufProcessor processor) {
+    public int forEachByte(ByteProcessor processor) {
         return -1;
     }
 
     @Override
-    public int forEachByte(int index, int length, ByteBufProcessor processor) {
+    public int forEachByte(int index, int length, ByteProcessor processor) {
         checkIndex(index, length);
         return -1;
     }
 
     @Override
-    public int forEachByteDesc(ByteBufProcessor processor) {
+    public int forEachByteDesc(ByteProcessor processor) {
         return -1;
     }
 
     @Override
-    public int forEachByteDesc(int index, int length, ByteBufProcessor processor) {
+    public int forEachByteDesc(int index, int length, ByteProcessor processor) {
         checkIndex(index, length);
         return -1;
     }
@@ -704,16 +870,31 @@ public final class EmptyByteBuf extends ByteBuf {
         return this;
     }
 
+    @Override
+    public ByteBuf retainedSlice() {
+        return this;
+    }
+
     @Override
     public ByteBuf slice(int index, int length) {
         return checkIndex(index, length);
     }
 
+    @Override
+    public ByteBuf retainedSlice(int index, int length) {
+        return checkIndex(index, length);
+    }
+
     @Override
     public ByteBuf duplicate() {
         return this;
     }
 
+    @Override
+    public ByteBuf retainedDuplicate() {
+        return this;
+    }
+
     @Override
     public int nioBufferCount() {
         return 1;
@@ -775,6 +956,11 @@ public final class EmptyByteBuf extends ByteBuf {
         }
     }
 
+    @Override
+    public boolean isContiguous() {
+        return true;
+    }
+
     @Override
     public String toString(Charset charset) {
         return "";
@@ -788,7 +974,7 @@ public final class EmptyByteBuf extends ByteBuf {
 
     @Override
     public int hashCode() {
-        return 0;
+        return EMPTY_BYTE_BUF_HASH_CODE;
     }
 
     @Override
@@ -831,6 +1017,16 @@ public final class EmptyByteBuf extends ByteBuf {
         return this;
     }
 
+    @Override
+    public ByteBuf touch() {
+        return this;
+    }
+
+    @Override
+    public ByteBuf touch(Object hint) {
+        return this;
+    }
+
     @Override
     public boolean release() {
         return false;
@@ -849,9 +1045,7 @@ public final class EmptyByteBuf extends ByteBuf {
     }
 
     private ByteBuf checkIndex(int index, int length) {
-        if (length < 0) {
-            throw new IllegalArgumentException("length: " + length);
-        }
+        checkPositiveOrZero(length, "length");
         if (index != 0 || length != 0) {
             throw new IndexOutOfBoundsException();
         }
@@ -859,9 +1053,7 @@ public final class EmptyByteBuf extends ByteBuf {
     }
 
     private ByteBuf checkLength(int length) {
-        if (length < 0) {
-            throw new IllegalArgumentException("length: " + length + " (expected: >= 0)");
-        }
+        checkPositiveOrZero(length, "length");
         if (length != 0) {
             throw new IndexOutOfBoundsException();
         }
