@@ -24,6 +24,7 @@ import java.io.OutputStream;
 import java.nio.ByteBuffer;
 import java.nio.ByteOrder;
 import java.nio.ReadOnlyBufferException;
+import java.nio.channels.FileChannel;
 import java.nio.channels.GatheringByteChannel;
 import java.nio.channels.ScatteringByteChannel;
 import java.util.Collections;
@@ -38,21 +39,20 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
     private final int capacity;
     private final ByteBufAllocator allocator;
     private final ByteOrder order;
-    private final Object[] buffers;
+    private final ByteBuf[] buffers;
     private final boolean direct;
 
     FixedCompositeByteBuf(ByteBufAllocator allocator, ByteBuf... buffers) {
-        super(Integer.MAX_VALUE);
+        super(AbstractByteBufAllocator.DEFAULT_MAX_CAPACITY);
         if (buffers.length == 0) {
             this.buffers = EMPTY;
             order = ByteOrder.BIG_ENDIAN;
             nioBufferCount = 1;
             capacity = 0;
-            direct = buffers[0].isDirect();
+            direct = Unpooled.EMPTY_BUFFER.isDirect();
         } else {
             ByteBuf b = buffers[0];
-            this.buffers = new Object[buffers.length];
-            this.buffers[0] = b;
+            this.buffers = buffers;
             boolean direct = true;
             int nioBufferCount = b.nioBufferCount();
             int capacity = b.readableBytes();
@@ -67,7 +67,6 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
                 if (!b.isDirect()) {
                     direct = false;
                 }
-                this.buffers[i] = b;
             }
             this.nioBufferCount = nioBufferCount;
             this.capacity = capacity;
@@ -127,6 +126,11 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
         throw new ReadOnlyBufferException();
     }
 
+    @Override
+    protected void _setShortLE(int index, int value) {
+        throw new ReadOnlyBufferException();
+    }
+
     @Override
     public ByteBuf setMedium(int index, int value) {
         throw new ReadOnlyBufferException();
@@ -137,6 +141,11 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
         throw new ReadOnlyBufferException();
     }
 
+    @Override
+    protected void _setMediumLE(int index, int value) {
+        throw new ReadOnlyBufferException();
+    }
+
     @Override
     public ByteBuf setInt(int index, int value) {
         throw new ReadOnlyBufferException();
@@ -147,6 +156,11 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
         throw new ReadOnlyBufferException();
     }
 
+    @Override
+    protected void _setIntLE(int index, int value) {
+        throw new ReadOnlyBufferException();
+    }
+
     @Override
     public ByteBuf setLong(int index, long value) {
         throw new ReadOnlyBufferException();
@@ -157,6 +171,11 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
         throw new ReadOnlyBufferException();
     }
 
+    @Override
+    protected void _setLongLE(int index, long value) {
+        throw new ReadOnlyBufferException();
+    }
+
     @Override
     public int setBytes(int index, InputStream in, int length) {
         throw new ReadOnlyBufferException();
@@ -167,6 +186,11 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
         throw new ReadOnlyBufferException();
     }
 
+    @Override
+    public int setBytes(int index, FileChannel in, long position, int length) {
+        throw new ReadOnlyBufferException();
+    }
+
     @Override
     public int capacity() {
         return capacity;
@@ -206,21 +230,15 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
         int readable = 0;
         for (int i = 0 ; i < buffers.length; i++) {
             Component comp = null;
-            ByteBuf b;
-            Object obj = buffers[i];
-            boolean isBuffer;
-            if (obj instanceof ByteBuf) {
-                b = (ByteBuf) obj;
-                isBuffer = true;
-            } else {
-                comp = (Component) obj;
+            ByteBuf b = buffers[i];
+            if (b instanceof Component) {
+                comp = (Component) b;
                 b = comp.buf;
-                isBuffer = false;
             }
             readable += b.readableBytes();
             if (index < readable) {
-                if (isBuffer) {
-                    // Create a new component ad store ti in the array so it not create a new object
+                if (comp == null) {
+                    // Create a new component and store it in the array so it not create a new object
                     // on the next access.
                     comp = new Component(i, readable - b.readableBytes(), b);
                     buffers[i] = comp;
@@ -235,11 +253,8 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
      * Return the {@link ByteBuf} stored at the given index of the array.
      */
     private ByteBuf buffer(int i) {
-        Object obj = buffers[i];
-        if (obj instanceof ByteBuf) {
-            return (ByteBuf) obj;
-        }
-        return ((Component) obj).buf;
+        ByteBuf b = buffers[i];
+        return b instanceof Component ? ((Component) b).buf : b;
     }
 
     @Override
@@ -265,6 +280,18 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
         }
     }
 
+    @Override
+    protected short _getShortLE(int index) {
+        Component c = findComponent(index);
+        if (index + 2 <= c.endOffset) {
+            return c.buf.getShortLE(index - c.offset);
+        } else if (order() == ByteOrder.BIG_ENDIAN) {
+            return (short) (_getByte(index) & 0xff | (_getByte(index + 1) & 0xff) << 8);
+        } else {
+            return (short) ((_getByte(index) & 0xff) << 8 | _getByte(index + 1) & 0xff);
+        }
+    }
+
     @Override
     protected int _getUnsignedMedium(int index) {
         Component c = findComponent(index);
@@ -277,6 +304,18 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
         }
     }
 
+    @Override
+    protected int _getUnsignedMediumLE(int index) {
+        Component c = findComponent(index);
+        if (index + 3 <= c.endOffset) {
+            return c.buf.getUnsignedMediumLE(index - c.offset);
+        } else if (order() == ByteOrder.BIG_ENDIAN) {
+            return _getShortLE(index) & 0xffff | (_getByte(index + 2) & 0xff) << 16;
+        } else {
+            return (_getShortLE(index) & 0xffff) << 8 | _getByte(index + 2) & 0xff;
+        }
+    }
+
     @Override
     protected int _getInt(int index) {
         Component c = findComponent(index);
@@ -289,6 +328,18 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
         }
     }
 
+    @Override
+    protected int _getIntLE(int index) {
+        Component c = findComponent(index);
+        if (index + 4 <= c.endOffset) {
+            return c.buf.getIntLE(index - c.offset);
+        } else if (order() == ByteOrder.BIG_ENDIAN) {
+            return _getShortLE(index) & 0xFFFF | (_getShortLE(index + 2) & 0xFFFF) << 16;
+        } else {
+            return (_getShortLE(index) & 0xffff) << 16 | _getShortLE(index + 2) & 0xffff;
+        }
+    }
+
     @Override
     protected long _getLong(int index) {
         Component c = findComponent(index);
@@ -301,6 +352,18 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
         }
     }
 
+    @Override
+    protected long _getLongLE(int index) {
+        Component c = findComponent(index);
+        if (index + 8 <= c.endOffset) {
+            return c.buf.getLongLE(index - c.offset);
+        } else if (order() == ByteOrder.BIG_ENDIAN) {
+            return _getIntLE(index) & 0xffffffffL | (_getIntLE(index + 4) & 0xffffffffL) << 32;
+        } else {
+            return (_getIntLE(index) & 0xffffffffL) << 32 | _getIntLE(index + 4) & 0xffffffffL;
+        }
+    }
+
     @Override
     public ByteBuf getBytes(int index, byte[] dst, int dstIndex, int length) {
         checkDstIndex(index, length, dstIndex, dst.length);
@@ -402,6 +465,25 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
         }
     }
 
+    @Override
+    public int getBytes(int index, FileChannel out, long position, int length)
+            throws IOException {
+        int count = nioBufferCount();
+        if (count == 1) {
+            return out.write(internalNioBuffer(index, length), position);
+        } else {
+            long writtenBytes = 0;
+            for (ByteBuffer buf : nioBuffers(index, length)) {
+                writtenBytes += out.write(buf, position + writtenBytes);
+            }
+            if (writtenBytes > Integer.MAX_VALUE) {
+                return Integer.MAX_VALUE;
+            } else {
+                return (int) writtenBytes;
+            }
+        }
+    }
+
     @Override
     public ByteBuf getBytes(int index, OutputStream out, int length) throws IOException {
         checkIndex(index, length);
@@ -511,7 +593,7 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
                 s = buffer(++i);
             }
 
-            return array.toArray(new ByteBuffer[array.size()]);
+            return array.toArray(new ByteBuffer[0]);
         } finally {
             array.recycle();
         }
@@ -519,27 +601,62 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
 
     @Override
     public boolean hasArray() {
-        return false;
+        switch (buffers.length) {
+            case 0:
+                return true;
+            case 1:
+                return buffer(0).hasArray();
+            default:
+                return false;
+        }
     }
 
     @Override
     public byte[] array() {
-        throw new UnsupportedOperationException();
+        switch (buffers.length) {
+            case 0:
+                return EmptyArrays.EMPTY_BYTES;
+            case 1:
+                return buffer(0).array();
+            default:
+                throw new UnsupportedOperationException();
+        }
     }
 
     @Override
     public int arrayOffset() {
-        throw new UnsupportedOperationException();
+        switch (buffers.length) {
+            case 0:
+                return 0;
+            case 1:
+                return buffer(0).arrayOffset();
+            default:
+                throw new UnsupportedOperationException();
+        }
     }
 
     @Override
     public boolean hasMemoryAddress() {
-        return false;
+        switch (buffers.length) {
+            case 0:
+                return Unpooled.EMPTY_BUFFER.hasMemoryAddress();
+            case 1:
+                return buffer(0).hasMemoryAddress();
+            default:
+                return false;
+        }
     }
 
     @Override
     public long memoryAddress() {
-        throw new UnsupportedOperationException();
+        switch (buffers.length) {
+            case 0:
+                return Unpooled.EMPTY_BUFFER.memoryAddress();
+            case 1:
+                return buffer(0).memoryAddress();
+            default:
+                throw new UnsupportedOperationException();
+        }
     }
 
     @Override
@@ -556,17 +673,16 @@ final class FixedCompositeByteBuf extends AbstractReferenceCountedByteBuf {
         return result + ", components=" + buffers.length + ')';
     }
 
-    private static final class Component {
+    private static final class Component extends WrappedByteBuf {
         private final int index;
         private final int offset;
-        private final ByteBuf buf;
         private final int endOffset;
 
         Component(int index, int offset, ByteBuf buf) {
+            super(buf);
             this.index = index;
             this.offset = offset;
             endOffset = offset + buf.readableBytes();
-            this.buf = buf;
         }
     }
 }
