@@ -16,20 +16,16 @@
 
 package io.netty.buffer;
 
-import io.netty.util.internal.ThreadLocalRandom;
 
-import static org.junit.Assert.assertEquals;
+import org.junit.Assert;
 
 public class RetainedSlicedByteBufTest extends SlicedByteBufTest {
-    @Override
-    protected ByteBuf newBuffer(int length) {
-        ByteBuf wrapped = Unpooled.wrappedBuffer(new byte[length * 2]);
-        ByteBuf buffer = wrapped.retainedSlice(length > 1 ? ThreadLocalRandom.current().nextInt(length - 1) + 1 : 0,
-                                               length);
-        wrapped.release();
 
-        assertEquals(0, buffer.readerIndex());
-        assertEquals(length, buffer.writerIndex());
-        return buffer;
+    @Override
+    protected ByteBuf newSlice(ByteBuf buffer, int offset, int length) {
+        ByteBuf slice = buffer.retainedSlice(offset, length);
+        buffer.release();
+        Assert.assertEquals(buffer.refCnt(), slice.refCnt());
+        return slice;
     }
 }
