@@ -15,27 +15,40 @@
  */
 package io.netty.buffer;
 
-import org.junit.Assert;
 import org.junit.Test;
 
+import static io.netty.buffer.Unpooled.buffer;
+import static org.junit.Assert.assertEquals;
+import static org.junit.Assert.assertFalse;
+import static org.junit.Assert.assertSame;
+import static org.junit.Assert.assertTrue;
+
 public class UnreleaseableByteBufTest {
 
     @Test
     public void testCantRelease() {
         ByteBuf buf = Unpooled.unreleasableBuffer(Unpooled.copyInt(1));
-        Assert.assertEquals(1, buf.refCnt());
-        Assert.assertFalse(buf.release());
-        Assert.assertEquals(1, buf.refCnt());
-        Assert.assertFalse(buf.release());
-        Assert.assertEquals(1, buf.refCnt());
+        assertEquals(1, buf.refCnt());
+        assertFalse(buf.release());
+        assertEquals(1, buf.refCnt());
+        assertFalse(buf.release());
+        assertEquals(1, buf.refCnt());
 
         buf.retain(5);
-        Assert.assertEquals(1, buf.refCnt());
+        assertEquals(1, buf.refCnt());
 
         buf.retain();
-        Assert.assertEquals(1, buf.refCnt());
+        assertEquals(1, buf.refCnt());
 
-        Assert.assertTrue(buf.unwrap().release());
-        Assert.assertEquals(0, buf.refCnt());
+        assertTrue(buf.unwrap().release());
+        assertEquals(0, buf.refCnt());
+    }
+
+    @Test
+    public void testWrappedReadOnly() {
+        ByteBuf buf = Unpooled.unreleasableBuffer(buffer(1).asReadOnly());
+        assertSame(buf, buf.asReadOnly());
+
+        assertTrue(buf.unwrap().release());
     }
 }
