@@ -15,13 +15,10 @@
  */
 package io.netty.handler.codec.rtsp;
 
-import static io.netty.handler.codec.http.HttpConstants.CR;
-import static io.netty.handler.codec.http.HttpConstants.LF;
-import static io.netty.handler.codec.http.HttpConstants.SP;
 import io.netty.buffer.ByteBuf;
+import io.netty.buffer.ByteBufUtil;
 import io.netty.handler.codec.UnsupportedMessageTypeException;
 import io.netty.handler.codec.http.HttpContent;
-import io.netty.handler.codec.http.HttpHeaders;
 import io.netty.handler.codec.http.HttpMessage;
 import io.netty.handler.codec.http.HttpObjectEncoder;
 import io.netty.handler.codec.http.HttpRequest;
@@ -29,15 +26,14 @@ import io.netty.handler.codec.http.HttpResponse;
 import io.netty.util.CharsetUtil;
 import io.netty.util.internal.StringUtil;
 
+import static io.netty.handler.codec.http.HttpConstants.*;
+
 /**
  * Encodes an RTSP message represented in {@link HttpMessage} or an {@link HttpContent} into
  * a {@link ByteBuf}.
  */
 public class RtspEncoder extends HttpObjectEncoder<HttpMessage> {
-    /**
-     * Constant for CRLF.
-     */
-    private static final byte[] CRLF = {CR, LF};
+    private static final int CRLF_SHORT = (CR << 8) | LF;
 
     @Override
     public boolean acceptOutboundMessage(final Object msg)
@@ -50,23 +46,20 @@ public class RtspEncoder extends HttpObjectEncoder<HttpMessage> {
            throws Exception {
         if (message instanceof HttpRequest) {
             HttpRequest request = (HttpRequest) message;
-            HttpHeaders.encodeAscii(request.method().toString(), buf);
+            ByteBufUtil.copy(request.method().asciiName(), buf);
             buf.writeByte(SP);
-            buf.writeBytes(request.uri().getBytes(CharsetUtil.UTF_8));
+            buf.writeCharSequence(request.uri(), CharsetUtil.UTF_8);
             buf.writeByte(SP);
-            HttpHeaders.encodeAscii(request.protocolVersion().toString(), buf);
-            buf.writeBytes(CRLF);
+            buf.writeCharSequence(request.protocolVersion().toString(), CharsetUtil.US_ASCII);
+            ByteBufUtil.writeShortBE(buf, CRLF_SHORT);
         } else if (message instanceof HttpResponse) {
             HttpResponse response = (HttpResponse) message;
-            HttpHeaders.encodeAscii(response.protocolVersion().toString(),
-                                    buf);
+            buf.writeCharSequence(response.protocolVersion().toString(), CharsetUtil.US_ASCII);
             buf.writeByte(SP);
-            buf.writeBytes(String.valueOf(response.status().code())
-                                 .getBytes(CharsetUtil.US_ASCII));
+            ByteBufUtil.copy(response.status().codeAsText(), buf);
             buf.writeByte(SP);
-            HttpHeaders.encodeAscii(String.valueOf(response.status().reasonPhrase()),
-                                    buf);
-            buf.writeBytes(CRLF);
+            buf.writeCharSequence(response.status().reasonPhrase(), CharsetUtil.US_ASCII);
+            ByteBufUtil.writeShortBE(buf, CRLF_SHORT);
         } else {
             throw new UnsupportedMessageTypeException("Unsupported type "
                                 + StringUtil.simpleClassName(message));
