@@ -21,6 +21,7 @@ import io.netty.util.CharsetUtil;
 import org.junit.Test;
 
 import java.net.IDN;
+import java.nio.CharBuffer;
 
 import static org.junit.Assert.*;
 
@@ -101,7 +102,7 @@ public class SocksCmdRequestTest {
     @Test
     public void testIDNEncodeToAsciiForDomain() {
         String host = "тест.рф";
-        String asciiHost = IDN.toASCII(host);
+        CharBuffer asciiHost = CharBuffer.wrap(IDN.toASCII(host));
         short port = 10000;
 
         SocksCmdRequest rq = new SocksCmdRequest(SocksCmdType.BIND, SocksAddressType.DOMAIN, host, port);
@@ -116,7 +117,8 @@ public class SocksCmdRequestTest {
         assertEquals((byte) 0x00, buffer.readByte());
         assertEquals(SocksAddressType.DOMAIN.byteValue(), buffer.readByte());
         assertEquals((byte) asciiHost.length(), buffer.readUnsignedByte());
-        assertEquals(asciiHost, buffer.readCharSequence(asciiHost.length(), CharsetUtil.US_ASCII));
+        assertEquals(asciiHost,
+            CharBuffer.wrap(buffer.readCharSequence(asciiHost.length(), CharsetUtil.US_ASCII)));
         assertEquals(port, buffer.readUnsignedShort());
 
         buffer.release();
