@@ -16,15 +16,14 @@
 package io.netty.handler.codec.compression;
 
 import io.netty.buffer.ByteBuf;
-import io.netty.buffer.ByteBufUtil;
 
 /**
  * Uncompresses an input {@link ByteBuf} encoded with Snappy compression into an
  * output {@link ByteBuf}.
  *
- * See http://code.google.com/p/snappy/source/browse/trunk/format_description.txt
+ * See <a href="https://github.com/google/snappy/blob/master/format_description.txt">snappy format</a>.
  */
-class Snappy {
+public final class Snappy {
 
     private static final int MAX_HT_SIZE = 1 << 14;
     private static final int MIN_COMPRESSIBLE_BYTES = 15;
@@ -73,7 +72,7 @@ class Snappy {
         final int baseIndex = inIndex;
 
         final short[] table = getHashTable(length);
-        final int shift = 32 - (int) Math.floor(Math.log(table.length) / Math.log(2));
+        final int shift = Integer.numberOfLeadingZeros(table.length) + 1;
 
         int nextEmit = inIndex;
 
@@ -149,7 +148,7 @@ class Snappy {
      * @return A 32-bit hash of 4 bytes located at index
      */
     private static int hash(ByteBuf in, int index, int shift) {
-        return in.getInt(index) + 0x1e35a7bd >>> shift;
+        return in.getInt(index) * 0x1e35a7bd >>> shift;
     }
 
     /**
@@ -163,15 +162,7 @@ class Snappy {
         while (htSize < MAX_HT_SIZE && htSize < inputSize) {
             htSize <<= 1;
         }
-
-        short[] table;
-        if (htSize <= 256) {
-            table = new short[256];
-        } else {
-            table = new short[MAX_HT_SIZE];
-        }
-
-        return table;
+        return new short[htSize];
     }
 
     /**
@@ -281,6 +272,7 @@ class Snappy {
             switch (state) {
             case READY:
                 state = State.READING_PREAMBLE;
+                // fall through
             case READING_PREAMBLE:
                 int uncompressedLength = readPreamble(in);
                 if (uncompressedLength == PREAMBLE_NOT_FULL) {
@@ -294,6 +286,7 @@ class Snappy {
                 }
                 out.ensureWritable(uncompressedLength);
                 state = State.READING_TAG;
+                // fall through
             case READING_TAG:
                 if (!in.isReadable()) {
                     return;
@@ -410,19 +403,19 @@ class Snappy {
             if (in.readableBytes() < 2) {
                 return NOT_ENOUGH_INPUT;
             }
-            length = ByteBufUtil.swapShort(in.readShort());
+            length = in.readUnsignedShortLE();
             break;
         case 62:
             if (in.readableBytes() < 3) {
                 return NOT_ENOUGH_INPUT;
             }
-            length = ByteBufUtil.swapMedium(in.readUnsignedMedium());
+            length = in.readUnsignedMediumLE();
             break;
         case 63:
             if (in.readableBytes() < 4) {
                 return NOT_ENOUGH_INPUT;
             }
-            length = ByteBufUtil.swapInt(in.readInt());
+            length = in.readIntLE();
             break;
         default:
             length = tag >> 2 & 0x3F;
@@ -502,7 +495,7 @@ class Snappy {
 
         int initialIndex = out.writerIndex();
         int length = 1 + (tag >> 2 & 0x03f);
-        int offset = ByteBufUtil.swapShort(in.readShort());
+        int offset = in.readUnsignedShortLE();
 
         validateOffset(offset, writtenSoFar);
 
@@ -546,7 +539,7 @@ class Snappy {
 
         int initialIndex = out.writerIndex();
         int length = 1 + (tag >> 2 & 0x03F);
-        int offset = ByteBufUtil.swapInt(in.readInt());
+        int offset = in.readIntLE();
 
         validateOffset(offset, writtenSoFar);
 
@@ -572,7 +565,7 @@ class Snappy {
 
     /**
      * Validates that the offset extracted from a compressed reference is within
-     * the permissible bounds of an offset (4 <= offset <= 32768), and does not
+     * the permissible bounds of an offset (0 < offset < Integer.MAX_VALUE), and does not
      * exceed the length of the chunk currently read so far.
      *
      * @param offset The offset extracted from the compressed reference
@@ -580,12 +573,13 @@ class Snappy {
      * @throws DecompressionException if the offset is invalid
      */
     private static void validateOffset(int offset, int chunkSizeSoFar) {
-        if (offset > Short.MAX_VALUE) {
-            throw new DecompressionException("Offset exceeds maximum permissible value");
+        if (offset == 0) {
+            throw new DecompressionException("Offset is less than minimum permissible value");
         }
 
-        if (offset <= 0) {
-            throw new DecompressionException("Offset is less than minimum permissible value");
+        if (offset < 0) {
+            // Due to arithmetic overflow
+            throw new DecompressionException("Offset is greater than maximum value supported by this implementation");
         }
 
         if (offset > chunkSizeSoFar) {
@@ -599,7 +593,7 @@ class Snappy {
      *
      * @param data The input data to calculate the CRC32C checksum of
      */
-    public static int calculateChecksum(ByteBuf data) {
+    static int calculateChecksum(ByteBuf data) {
         return calculateChecksum(data, data.readerIndex(), data.readableBytes());
     }
 
@@ -609,18 +603,11 @@ class Snappy {
      *
      * @param data The input data to calculate the CRC32C checksum of
      */
-    public static int calculateChecksum(ByteBuf data, int offset, int length) {
+    static int calculateChecksum(ByteBuf data, int offset, int length) {
         Crc32c crc32 = new Crc32c();
         try {
-            if (data.hasArray()) {
-                crc32.update(data.array(), data.arrayOffset() + offset, length);
-            } else {
-                byte[] array = new byte[length];
-                data.getBytes(offset, array);
-                crc32.update(array, 0, length);
-            }
-
-            return maskChecksum((int) crc32.getValue());
+            crc32.update(data, offset, length);
+            return maskChecksum(crc32.getValue());
         } finally {
             crc32.reset();
         }
@@ -668,7 +655,7 @@ class Snappy {
      * @param checksum The actual checksum of the data
      * @return The masked checksum
      */
-    static int maskChecksum(int checksum) {
-        return (checksum >> 15 | checksum << 17) + 0xa282ead8;
+    static int maskChecksum(long checksum) {
+        return (int) ((checksum >> 15 | checksum << 17) + 0xa282ead8);
     }
 }
