@@ -51,8 +51,17 @@ public class DatagramPacketEncoderTest {
 
     @Test
     public void testEncode() {
+        testEncode(false);
+    }
+
+    @Test
+    public void testEncodeWithSenderIsNull() {
+        testEncode(true);
+    }
+
+    private void testEncode(boolean senderIsNull) {
         InetSocketAddress recipient = SocketUtils.socketAddress("127.0.0.1", 10000);
-        InetSocketAddress sender = SocketUtils.socketAddress("127.0.0.1", 20000);
+        InetSocketAddress sender = senderIsNull ? null : SocketUtils.socketAddress("127.0.0.1", 20000);
         assertTrue(channel.writeOutbound(
                 new DefaultAddressedEnvelope<String, InetSocketAddress>("netty", recipient, sender)));
         DatagramPacket packet = channel.readOutbound();
