@@ -176,11 +176,6 @@ class PromiseTask<V> extends DefaultPromise<V> implements RunnableFuture<V> {
         return clearTaskAfterCompletion(super.cancel(mayInterruptIfRunning), CANCELLED);
     }
 
-    @Override
-    public final boolean isCancelled() {
-        return task == CANCELLED || super.isCancelled();
-    }
-
     @Override
     protected StringBuilder toStringBuilder() {
         StringBuilder buf = super.toStringBuilder();
