@@ -15,20 +15,21 @@
  */
 package io.netty.util.internal;
 
-import io.netty.util.Recycler;
 import io.netty.util.ReferenceCountUtil;
 import io.netty.util.concurrent.Promise;
+import io.netty.util.internal.ObjectPool.Handle;
+import io.netty.util.internal.ObjectPool.ObjectCreator;
 
 /**
  * Some pending write which should be picked up later.
  */
 public final class PendingWrite {
-    private static final Recycler<PendingWrite> RECYCLER = new Recycler<PendingWrite>() {
+    private static final ObjectPool<PendingWrite> RECYCLER = ObjectPool.newPool(new ObjectCreator<PendingWrite>() {
         @Override
-        protected PendingWrite newObject(Handle<PendingWrite> handle) {
+        public PendingWrite newObject(Handle<PendingWrite> handle) {
             return new PendingWrite(handle);
         }
-    };
+    });
 
     /**
      * Create a new empty {@link RecyclableArrayList} instance
@@ -40,11 +41,11 @@ public final class PendingWrite {
         return pending;
     }
 
-    private final Recycler.Handle<PendingWrite> handle;
+    private final Handle<PendingWrite> handle;
     private Object msg;
     private Promise<Void> promise;
 
-    private PendingWrite(Recycler.Handle<PendingWrite> handle) {
+    private PendingWrite(Handle<PendingWrite> handle) {
         this.handle = handle;
     }
 
@@ -59,7 +60,7 @@ public final class PendingWrite {
     }
 
     /**
-     * Fails the underlying {@link Promise} with the given cause and reycle this instance.
+     * Fails the underlying {@link Promise} with the given cause and recycle this instance.
      */
     public boolean failAndRecycle(Throwable cause) {
         ReferenceCountUtil.release(msg);
@@ -70,7 +71,7 @@ public final class PendingWrite {
     }
 
     /**
-     * Mark the underlying {@link Promise} successed and reycle this instance.
+     * Mark the underlying {@link Promise} successfully and recycle this instance.
      */
     public boolean successAndRecycle() {
         if (promise != null) {
