@@ -1,5 +1,5 @@
 /*
- * Copyright 2015 The Netty Project
+ * Copyright 2017 The Netty Project
  *
  * The Netty Project licenses this file to you under the Apache License,
  * version 2.0 (the "License"); you may not use this file except in compliance
@@ -17,7 +17,13 @@ package io.netty.handler.ssl;
 
 import org.junit.BeforeClass;
 
-import static org.junit.Assume.assumeFalse;
+import java.util.concurrent.atomic.AtomicReference;
+
+import javax.net.ssl.SSLException;
+
+import static org.hamcrest.MatcherAssert.assertThat;
+import static org.hamcrest.Matchers.instanceOf;
+import static org.hamcrest.Matchers.is;
 import static org.junit.Assume.assumeTrue;
 
 public class OpenSslRenegotiateTest extends RenegotiateTest {
@@ -25,12 +31,16 @@ public class OpenSslRenegotiateTest extends RenegotiateTest {
     @BeforeClass
     public static void checkOpenSsl() {
         assumeTrue(OpenSsl.isAvailable());
-        // BoringSSL does not support renegotiation intentionally.
-        assumeFalse("BoringSSL".equals(OpenSsl.versionString()));
     }
 
     @Override
     protected SslProvider serverSslProvider() {
         return SslProvider.OPENSSL;
     }
+
+    protected void verifyResult(AtomicReference<Throwable> error) throws Throwable {
+        Throwable cause = error.get();
+        // Renegotation is not supported by the OpenSslEngine.
+        assertThat(cause, is(instanceOf(SSLException.class)));
+    }
 }
