@@ -16,7 +16,6 @@
 package io.netty.resolver.dns;
 
 import io.netty.handler.codec.dns.DnsQuestion;
-import io.netty.util.internal.UnstableApi;
 
 import java.net.InetSocketAddress;
 
@@ -24,7 +23,6 @@ import java.net.InetSocketAddress;
  * A {@link DnsNameResolverException} raised when {@link DnsNameResolver} failed to perform a successful query because
  * of an timeout. In this case you may want to retry the operation.
  */
-@UnstableApi
 public final class DnsNameResolverTimeoutException extends DnsNameResolverException {
     private static final long serialVersionUID = -8826717969627131854L;
 
