@@ -17,7 +17,6 @@ package io.netty.resolver.dns;
 
 import io.netty.util.NetUtil;
 import io.netty.util.internal.SocketUtils;
-import io.netty.util.internal.UnstableApi;
 import io.netty.util.internal.logging.InternalLogger;
 import io.netty.util.internal.logging.InternalLoggerFactory;
 
@@ -44,7 +43,6 @@ import static io.netty.util.internal.StringUtil.indexOfWhiteSpace;
  * <a href="https://developer.apple.com/legacy/library/documentation/Darwin/Reference/ManPages/man5/resolver.5.html">
  * /etc/resolver</a> to respect the system default domain servers.
  */
-@UnstableApi
 public final class UnixResolverDnsServerAddressStreamProvider implements DnsServerAddressStreamProvider {
     private static final InternalLogger logger =
             InternalLoggerFactory.getInstance(UnixResolverDnsServerAddressStreamProvider.class);
@@ -73,7 +71,9 @@ public final class UnixResolverDnsServerAddressStreamProvider implements DnsServ
             return nameServerCache.mayOverrideNameServers() ? nameServerCache
                                                             : DefaultDnsServerAddressStreamProvider.INSTANCE;
         } catch (Exception e) {
-            logger.debug("failed to parse {} and/or {}", ETC_RESOLV_CONF_FILE, ETC_RESOLVER_DIR, e);
+            if (logger.isDebugEnabled()) {
+                logger.debug("failed to parse {} and/or {}", ETC_RESOLV_CONF_FILE, ETC_RESOLVER_DIR, e);
+            }
             return DefaultDnsServerAddressStreamProvider.INSTANCE;
         }
     }
@@ -256,8 +256,10 @@ public final class UnixResolverDnsServerAddressStreamProvider implements DnsServ
         DnsServerAddresses existingAddresses = domainToNameServerStreamMap.put(domainName, addresses);
         if (existingAddresses != null) {
             domainToNameServerStreamMap.put(domainName, existingAddresses);
-            logger.debug("Domain name {} already maps to addresses {} so new addresses {} will be discarded",
-                    domainName, existingAddresses, addresses);
+            if (logger.isDebugEnabled()) {
+                logger.debug("Domain name {} already maps to addresses {} so new addresses {} will be discarded",
+                        domainName, existingAddresses, addresses);
+            }
         }
     }
 
