@@ -18,7 +18,4 @@
  * An alternative to Java's built-in domain name lookup mechanism that resolves a domain name asynchronously,
  * which supports the queries of an arbitrary DNS record type as well.
  */
-@UnstableApi
 package io.netty.resolver.dns;
-
-import io.netty.util.internal.UnstableApi;
