@@ -164,6 +164,19 @@ public class UnixResolverDnsServerAddressStreamProviderTest {
         assertEquals(Collections.singletonList("squarecorp.local"), domains);
     }
 
+    @Test
+    public void ignoreInvalidEntries() throws Exception {
+        File f = buildFile("domain netty.local\n" +
+                "nameserver nil\n" +
+                "nameserver 127.0.0.3\n");
+        UnixResolverDnsServerAddressStreamProvider p =
+                new UnixResolverDnsServerAddressStreamProvider(f, null);
+
+        DnsServerAddressStream stream = p.nameServerAddressStream("somehost");
+        assertEquals(1, stream.size());
+        assertHostNameEquals("127.0.0.3", stream.next());
+    }
+
     private File buildFile(String contents) throws IOException {
         File f = folder.newFile();
         OutputStream out = new FileOutputStream(f);
@@ -175,6 +188,17 @@ public class UnixResolverDnsServerAddressStreamProviderTest {
         return f;
     }
 
+    @Test
+    public void ignoreComments() throws Exception {
+        File f = buildFile("domain linecorp.local\n" +
+                "nameserver 127.0.0.2 #somecomment\n");
+        UnixResolverDnsServerAddressStreamProvider p =
+                new UnixResolverDnsServerAddressStreamProvider(f, null);
+
+        DnsServerAddressStream stream = p.nameServerAddressStream("somehost");
+        assertHostNameEquals("127.0.0.2", stream.next());
+    }
+
     private static void assertHostNameEquals(String expectedHostname, InetSocketAddress next) {
         assertEquals("unexpected hostname: " + next, expectedHostname, next.getHostString());
     }
