@@ -20,13 +20,17 @@ import static org.junit.Assume.*;
 import io.netty.bootstrap.Bootstrap;
 import io.netty.channel.ChannelException;
 import io.netty.channel.ChannelInboundHandlerAdapter;
+import io.netty.channel.ChannelOption;
 import io.netty.channel.EventLoopGroup;
 
 import java.net.InetSocketAddress;
 import java.nio.channels.ClosedChannelException;
+import java.util.Map;
 import java.util.Random;
 
+import org.junit.After;
 import org.junit.AfterClass;
+import org.junit.Before;
 import org.junit.BeforeClass;
 import org.junit.Test;
 
@@ -37,9 +41,18 @@ public class EpollSocketChannelConfigTest {
     private static Random rand;
 
     @BeforeClass
-    public static void before() {
+    public static void beforeClass() {
         rand = new Random();
         group = new EpollEventLoopGroup(1);
+    }
+
+    @AfterClass
+    public static void afterClass() {
+        group.shutdownGracefully();
+    }
+
+    @Before
+    public void setup() {
         Bootstrap bootstrap = new Bootstrap();
         ch = (EpollSocketChannel) bootstrap.group(group)
                 .channel(EpollSocketChannel.class)
@@ -47,20 +60,16 @@ public class EpollSocketChannelConfigTest {
                 .bind(new InetSocketAddress(0)).syncUninterruptibly().channel();
     }
 
-    @AfterClass
-    public static void after() {
-        try {
-            ch.close().syncUninterruptibly();
-        } finally {
-            group.shutdownGracefully();
-        }
+    @After
+    public void teardown() {
+        ch.close().syncUninterruptibly();
     }
 
-    private long randLong(long min, long max) {
+    private static long randLong(long min, long max) {
         return min + nextLong(max - min + 1);
     }
 
-    private long nextLong(long n) {
+    private static long nextLong(long n) {
         long bits, val;
         do {
            bits = (rand.nextLong() << 1) >>> 1;
@@ -140,10 +149,16 @@ public class EpollSocketChannelConfigTest {
     public void testGetOptionWhenClosed() {
         ch.close().syncUninterruptibly();
         try {
-        ch.config().getSoLinger();
+            ch.config().getSoLinger();
             fail();
         } catch (ChannelException e) {
             assertTrue(e.getCause() instanceof ClosedChannelException);
         }
     }
+
+    @Test
+    public void getGetOptions() {
+        Map<ChannelOption<?>, Object> map = ch.config().getOptions();
+        assertFalse(map.isEmpty());
+    }
 }
