@@ -15,6 +15,10 @@
  */
 package io.netty.channel;
 
+import io.netty.util.internal.ObjectUtil;
+
+import static io.netty.util.internal.ObjectUtil.checkPositiveOrZero;
+
 import java.util.ArrayDeque;
 import java.util.Queue;
 
@@ -61,12 +65,8 @@ public final class ChannelFlushPromiseNotifier {
      * {@code pendingDataSize} was reached.
      */
     public ChannelFlushPromiseNotifier add(ChannelPromise promise, long pendingDataSize) {
-        if (promise == null) {
-            throw new NullPointerException("promise");
-        }
-        if (pendingDataSize < 0) {
-            throw new IllegalArgumentException("pendingDataSize must be >= 0 but was " + pendingDataSize);
-        }
+        ObjectUtil.checkNotNull(promise, "promise");
+        checkPositiveOrZero(pendingDataSize, "pendingDataSize");
         long checkpoint = writeCounter + pendingDataSize;
         if (promise instanceof FlushCheckpoint) {
             FlushCheckpoint cp = (FlushCheckpoint) promise;
@@ -81,9 +81,7 @@ public final class ChannelFlushPromiseNotifier {
      * Increase the current write counter by the given delta
      */
     public ChannelFlushPromiseNotifier increaseWriteCounter(long delta) {
-        if (delta < 0) {
-            throw new IllegalArgumentException("delta must be >= 0 but was " + delta);
-        }
+        checkPositiveOrZero(delta, "delta");
         writeCounter += delta;
         return this;
     }
