@@ -23,10 +23,7 @@ final class DefaultChannelHandlerContext extends AbstractChannelHandlerContext {
 
     DefaultChannelHandlerContext(
             DefaultChannelPipeline pipeline, EventExecutor executor, String name, ChannelHandler handler) {
-        super(pipeline, executor, name, isInbound(handler), isOutbound(handler));
-        if (handler == null) {
-            throw new NullPointerException("handler");
-        }
+        super(pipeline, executor, name, handler.getClass());
         this.handler = handler;
     }
 
@@ -34,12 +31,4 @@ final class DefaultChannelHandlerContext extends AbstractChannelHandlerContext {
     public ChannelHandler handler() {
         return handler;
     }
-
-    private static boolean isInbound(ChannelHandler handler) {
-        return handler instanceof ChannelInboundHandler;
-    }
-
-    private static boolean isOutbound(ChannelHandler handler) {
-        return handler instanceof ChannelOutboundHandler;
-    }
 }
