@@ -20,12 +20,15 @@ import io.netty.channel.Channel;
 import io.netty.util.concurrent.EventExecutor;
 import io.netty.util.concurrent.Future;
 import io.netty.util.concurrent.FutureListener;
+import io.netty.util.concurrent.GlobalEventExecutor;
 import io.netty.util.concurrent.Promise;
-import io.netty.util.internal.ThrowableUtil;
+import io.netty.util.internal.ObjectUtil;
 
 import java.nio.channels.ClosedChannelException;
 import java.util.ArrayDeque;
 import java.util.Queue;
+import java.util.concurrent.atomic.AtomicInteger;
+import java.util.concurrent.Callable;
 import java.util.concurrent.ScheduledFuture;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.TimeoutException;
@@ -35,12 +38,6 @@ import java.util.concurrent.TimeoutException;
  * number of concurrent connections.
  */
 public class FixedChannelPool extends SimpleChannelPool {
-    private static final IllegalStateException FULL_EXCEPTION = ThrowableUtil.unknownStackTrace(
-            new IllegalStateException("Too many outstanding acquire operations"),
-            FixedChannelPool.class, "acquire0(...)");
-    private static final TimeoutException TIMEOUT_EXCEPTION = ThrowableUtil.unknownStackTrace(
-            new TimeoutException("Acquire operation took longer then configured maximum time"),
-            FixedChannelPool.class, "<init>(...)");
 
     public enum AcquireTimeoutAction {
         /**
@@ -63,7 +60,7 @@ public class FixedChannelPool extends SimpleChannelPool {
     private final Queue<AcquireTask> pendingAcquireQueue = new ArrayDeque<AcquireTask>();
     private final int maxConnections;
     private final int maxPendingAcquires;
-    private int acquiredChannelCount;
+    private final AtomicInteger acquiredChannelCount = new AtomicInteger();
     private int pendingAcquireCount;
     private boolean closed;
 
@@ -72,7 +69,7 @@ public class FixedChannelPool extends SimpleChannelPool {
      *
      * @param bootstrap         the {@link Bootstrap} that is used for connections
      * @param handler           the {@link ChannelPoolHandler} that will be notified for the different pool actions
-     * @param maxConnections    the numnber of maximal active connections, once this is reached new tries to acquire
+     * @param maxConnections    the number of maximal active connections, once this is reached new tries to acquire
      *                          a {@link Channel} will be delayed until a connection is returned to the pool again.
      */
     public FixedChannelPool(Bootstrap bootstrap,
@@ -85,7 +82,7 @@ public class FixedChannelPool extends SimpleChannelPool {
      *
      * @param bootstrap             the {@link Bootstrap} that is used for connections
      * @param handler               the {@link ChannelPoolHandler} that will be notified for the different pool actions
-     * @param maxConnections        the numnber of maximal active connections, once this is reached new tries to
+     * @param maxConnections        the number of maximal active connections, once this is reached new tries to
      *                              acquire a {@link Channel} will be delayed until a connection is returned to the
      *                              pool again.
      * @param maxPendingAcquires    the maximum number of pending acquires. Once this is exceed acquire tries will
@@ -102,12 +99,12 @@ public class FixedChannelPool extends SimpleChannelPool {
      * @param bootstrap             the {@link Bootstrap} that is used for connections
      * @param handler               the {@link ChannelPoolHandler} that will be notified for the different pool actions
      * @param healthCheck           the {@link ChannelHealthChecker} that will be used to check if a {@link Channel} is
-     *                              still healty when obtain from the {@link ChannelPool}
+     *                              still healthy when obtain from the {@link ChannelPool}
      * @param action                the {@link AcquireTimeoutAction} to use or {@code null} if non should be used.
      *                              In this case {@param acquireTimeoutMillis} must be {@code -1}.
      * @param acquireTimeoutMillis  the time (in milliseconds) after which an pending acquire must complete or
      *                              the {@link AcquireTimeoutAction} takes place.
-     * @param maxConnections        the numnber of maximal active connections, once this is reached new tries to
+     * @param maxConnections        the number of maximal active connections, once this is reached new tries to
      *                              acquire a {@link Channel} will be delayed until a connection is returned to the
      *                              pool again.
      * @param maxPendingAcquires    the maximum number of pending acquires. Once this is exceed acquire tries will
@@ -127,12 +124,12 @@ public class FixedChannelPool extends SimpleChannelPool {
      * @param bootstrap             the {@link Bootstrap} that is used for connections
      * @param handler               the {@link ChannelPoolHandler} that will be notified for the different pool actions
      * @param healthCheck           the {@link ChannelHealthChecker} that will be used to check if a {@link Channel} is
-     *                              still healty when obtain from the {@link ChannelPool}
+     *                              still healthy when obtain from the {@link ChannelPool}
      * @param action                the {@link AcquireTimeoutAction} to use or {@code null} if non should be used.
      *                              In this case {@param acquireTimeoutMillis} must be {@code -1}.
      * @param acquireTimeoutMillis  the time (in milliseconds) after which an pending acquire must complete or
      *                              the {@link AcquireTimeoutAction} takes place.
-     * @param maxConnections        the numnber of maximal active connections, once this is reached new tries to
+     * @param maxConnections        the number of maximal active connections, once this is reached new tries to
      *                              acquire a {@link Channel} will be delayed until a connection is returned to the
      *                              pool again.
      * @param maxPendingAcquires    the maximum number of pending acquires. Once this is exceed acquire tries will
@@ -145,7 +142,37 @@ public class FixedChannelPool extends SimpleChannelPool {
                             ChannelHealthChecker healthCheck, AcquireTimeoutAction action,
                             final long acquireTimeoutMillis,
                             int maxConnections, int maxPendingAcquires, final boolean releaseHealthCheck) {
-        super(bootstrap, handler, healthCheck, releaseHealthCheck);
+        this(bootstrap, handler, healthCheck, action, acquireTimeoutMillis, maxConnections, maxPendingAcquires,
+                releaseHealthCheck, true);
+    }
+
+    /**
+     * Creates a new instance.
+     *
+     * @param bootstrap             the {@link Bootstrap} that is used for connections
+     * @param handler               the {@link ChannelPoolHandler} that will be notified for the different pool actions
+     * @param healthCheck           the {@link ChannelHealthChecker} that will be used to check if a {@link Channel} is
+     *                              still healthy when obtain from the {@link ChannelPool}
+     * @param action                the {@link AcquireTimeoutAction} to use or {@code null} if non should be used.
+     *                              In this case {@param acquireTimeoutMillis} must be {@code -1}.
+     * @param acquireTimeoutMillis  the time (in milliseconds) after which an pending acquire must complete or
+     *                              the {@link AcquireTimeoutAction} takes place.
+     * @param maxConnections        the number of maximal active connections, once this is reached new tries to
+     *                              acquire a {@link Channel} will be delayed until a connection is returned to the
+     *                              pool again.
+     * @param maxPendingAcquires    the maximum number of pending acquires. Once this is exceed acquire tries will
+     *                              be failed.
+     * @param releaseHealthCheck    will check channel health before offering back if this parameter set to
+     *                              {@code true}.
+     * @param lastRecentUsed        {@code true} {@link Channel} selection will be LIFO, if {@code false} FIFO.
+     */
+    public FixedChannelPool(Bootstrap bootstrap,
+                            ChannelPoolHandler handler,
+                            ChannelHealthChecker healthCheck, AcquireTimeoutAction action,
+                            final long acquireTimeoutMillis,
+                            int maxConnections, int maxPendingAcquires,
+                            boolean releaseHealthCheck, boolean lastRecentUsed) {
+        super(bootstrap, handler, healthCheck, releaseHealthCheck, lastRecentUsed);
         if (maxConnections < 1) {
             throw new IllegalArgumentException("maxConnections: " + maxConnections + " (expected: >= 1)");
         }
@@ -158,7 +185,7 @@ public class FixedChannelPool extends SimpleChannelPool {
         } else if (action == null && acquireTimeoutMillis != -1) {
             throw new NullPointerException("action");
         } else if (action != null && acquireTimeoutMillis < 0) {
-            throw new IllegalArgumentException("acquireTimeoutMillis: " + acquireTimeoutMillis + " (expected: >= 1)");
+            throw new IllegalArgumentException("acquireTimeoutMillis: " + acquireTimeoutMillis + " (expected: >= 0)");
         } else {
             acquireTimeoutNanos = TimeUnit.MILLISECONDS.toNanos(acquireTimeoutMillis);
             switch (action) {
@@ -167,7 +194,13 @@ public class FixedChannelPool extends SimpleChannelPool {
                     @Override
                     public void onTimeout(AcquireTask task) {
                         // Fail the promise as we timed out.
-                        task.promise.setFailure(TIMEOUT_EXCEPTION);
+                        task.promise.setFailure(new TimeoutException(
+                                "Acquire operation took longer then configured maximum time") {
+                            @Override
+                            public Throwable fillInStackTrace() {
+                                return this;
+                            }
+                        });
                     }
                 };
                 break;
@@ -176,7 +209,7 @@ public class FixedChannelPool extends SimpleChannelPool {
                     @Override
                     public void onTimeout(AcquireTask task) {
                         // Increment the acquire count and delegate to super to actually acquire a Channel which will
-                        // create a new connetion.
+                        // create a new connection.
                         task.acquired();
 
                         FixedChannelPool.super.acquire(task.promise);
@@ -192,6 +225,11 @@ public class FixedChannelPool extends SimpleChannelPool {
         this.maxPendingAcquires = maxPendingAcquires;
     }
 
+    /** Returns the number of acquired channels that this pool thinks it has. */
+    public int acquiredChannelCount() {
+        return acquiredChannelCount.get();
+    }
+
     @Override
     public Future<Channel> acquire(final Promise<Channel> promise) {
         try {
@@ -215,11 +253,11 @@ public class FixedChannelPool extends SimpleChannelPool {
         assert executor.inEventLoop();
 
         if (closed) {
-            promise.setFailure(new IllegalStateException("FixedChannelPooled was closed"));
+            promise.setFailure(new IllegalStateException("FixedChannelPool was closed"));
             return;
         }
-        if (acquiredChannelCount < maxConnections) {
-            assert acquiredChannelCount >= 0;
+        if (acquiredChannelCount.get() < maxConnections) {
+            assert acquiredChannelCount.get() >= 0;
 
             // We need to create a new promise as we need to ensure the AcquireListener runs in the correct
             // EventLoop
@@ -230,7 +268,7 @@ public class FixedChannelPool extends SimpleChannelPool {
             super.acquire(p);
         } else {
             if (pendingAcquireCount >= maxPendingAcquires) {
-                promise.setFailure(FULL_EXCEPTION);
+                tooManyOutstanding(promise);
             } else {
                 AcquireTask task = new AcquireTask(promise);
                 if (pendingAcquireQueue.offer(task)) {
@@ -240,7 +278,7 @@ public class FixedChannelPool extends SimpleChannelPool {
                         task.timeoutFuture = executor.schedule(timeoutTask, acquireTimeoutNanos, TimeUnit.NANOSECONDS);
                     }
                 } else {
-                    promise.setFailure(FULL_EXCEPTION);
+                    tooManyOutstanding(promise);
                 }
             }
 
@@ -248,8 +286,13 @@ public class FixedChannelPool extends SimpleChannelPool {
         }
     }
 
+    private void tooManyOutstanding(Promise<?> promise) {
+        promise.setFailure(new IllegalStateException("Too many outstanding acquire operations"));
+    }
+
     @Override
     public Future<Void> release(final Channel channel, final Promise<Void> promise) {
+        ObjectUtil.checkNotNull(promise, "promise");
         final Promise<Void> p = executor.newPromise();
         super.release(channel, p.addListener(new FutureListener<Void>() {
 
@@ -258,7 +301,9 @@ public class FixedChannelPool extends SimpleChannelPool {
                 assert executor.inEventLoop();
 
                 if (closed) {
-                    promise.setFailure(new IllegalStateException("FixedChannelPooled was closed"));
+                    // Since the pool is closed, we have no choice but to close the channel
+                    channel.close();
+                    promise.setFailure(new IllegalStateException("FixedChannelPool was closed"));
                     return;
                 }
 
@@ -275,14 +320,13 @@ public class FixedChannelPool extends SimpleChannelPool {
                 }
             }
         }));
-        return p;
+        return promise;
     }
 
     private void decrementAndRunTaskQueue() {
-        --acquiredChannelCount;
-
         // We should never have a negative value.
-        assert acquiredChannelCount >= 0;
+        int currentCount = acquiredChannelCount.decrementAndGet();
+        assert currentCount >= 0;
 
         // Run the pending acquire tasks before notify the original promise so if the user would
         // try to acquire again from the ChannelFutureListener and the pendingAcquireCount is >=
@@ -292,7 +336,7 @@ public class FixedChannelPool extends SimpleChannelPool {
     }
 
     private void runTaskQueue() {
-        while (acquiredChannelCount < maxConnections) {
+        while (acquiredChannelCount.get() < maxConnections) {
             AcquireTask task = pendingAcquireQueue.poll();
             if (task == null) {
                 break;
@@ -312,7 +356,7 @@ public class FixedChannelPool extends SimpleChannelPool {
 
         // We should never have a negative value.
         assert pendingAcquireCount >= 0;
-        assert acquiredChannelCount >= 0;
+        assert acquiredChannelCount.get() >= 0;
     }
 
     // AcquireTask extends AcquireListener to reduce object creations and so GC pressure
@@ -321,7 +365,7 @@ public class FixedChannelPool extends SimpleChannelPool {
         final long expireNanoTime = System.nanoTime() + acquireTimeoutNanos;
         ScheduledFuture<?> timeoutFuture;
 
-        public AcquireTask(Promise<Channel> promise) {
+        AcquireTask(Promise<Channel> promise) {
             super(promise);
             // We need to create a new promise as we need to ensure the AcquireListener runs in the correct
             // EventLoop.
@@ -366,7 +410,11 @@ public class FixedChannelPool extends SimpleChannelPool {
             assert executor.inEventLoop();
 
             if (closed) {
-                originalPromise.setFailure(new IllegalStateException("FixedChannelPooled was closed"));
+                if (future.isSuccess()) {
+                    // Since the pool is closed, we have no choice but to close the channel
+                    future.getNow().close();
+                }
+                originalPromise.setFailure(new IllegalStateException("FixedChannelPool was closed"));
                 return;
             }
 
@@ -387,34 +435,81 @@ public class FixedChannelPool extends SimpleChannelPool {
             if (acquired) {
                 return;
             }
-            acquiredChannelCount++;
+            acquiredChannelCount.incrementAndGet();
             acquired = true;
         }
     }
 
     @Override
     public void close() {
-        executor.execute(new Runnable() {
-            @Override
-            public void run() {
-                if (!closed) {
-                    closed = true;
-                    for (;;) {
-                        AcquireTask task = pendingAcquireQueue.poll();
-                        if (task == null) {
-                            break;
+        try {
+            closeAsync().await();
+        } catch (InterruptedException e) {
+            Thread.currentThread().interrupt();
+            throw new RuntimeException(e);
+        }
+    }
+
+    /**
+     * Closes the pool in an async manner.
+     *
+     * @return Future which represents completion of the close task
+     */
+    @Override
+    public Future<Void> closeAsync() {
+        if (executor.inEventLoop()) {
+            return close0();
+        } else {
+            final Promise<Void> closeComplete = executor.newPromise();
+            executor.execute(new Runnable() {
+                @Override
+                public void run() {
+                    close0().addListener(new FutureListener<Void>() {
+                        @Override
+                        public void operationComplete(Future<Void> f) throws Exception {
+                            if (f.isSuccess()) {
+                                closeComplete.setSuccess(null);
+                            } else {
+                                closeComplete.setFailure(f.cause());
+                            }
                         }
-                        ScheduledFuture<?> f = task.timeoutFuture;
-                        if (f != null) {
-                            f.cancel(false);
-                        }
-                        task.promise.setFailure(new ClosedChannelException());
-                    }
-                    acquiredChannelCount = 0;
-                    pendingAcquireCount = 0;
-                    FixedChannelPool.super.close();
+                    });
                 }
+            });
+            return closeComplete;
+        }
+    }
+
+    private Future<Void> close0() {
+        assert executor.inEventLoop();
+
+        if (!closed) {
+            closed = true;
+            for (;;) {
+                AcquireTask task = pendingAcquireQueue.poll();
+                if (task == null) {
+                    break;
+                }
+                ScheduledFuture<?> f = task.timeoutFuture;
+                if (f != null) {
+                    f.cancel(false);
+                }
+                task.promise.setFailure(new ClosedChannelException());
             }
-        });
+            acquiredChannelCount.set(0);
+            pendingAcquireCount = 0;
+
+            // Ensure we dispatch this on another Thread as close0 will be called from the EventExecutor and we need
+            // to ensure we will not block in a EventExecutor.
+            return GlobalEventExecutor.INSTANCE.submit(new Callable<Void>() {
+                @Override
+                public Void call() throws Exception {
+                    FixedChannelPool.super.close();
+                    return null;
+                }
+            });
+        }
+
+        return GlobalEventExecutor.INSTANCE.newSucceededFuture(null);
     }
 }
