@@ -264,6 +264,30 @@ public class PendingWriteQueueTest {
         assertEquals(3L, channel.readOutbound());
     }
 
+    @Test
+    public void testRemoveAndWriteAllWithVoidPromise() {
+        EmbeddedChannel channel = new EmbeddedChannel(new ChannelOutboundHandlerAdapter() {
+            @Override
+            public void write(ChannelHandlerContext ctx, Object msg, ChannelPromise promise) {
+                // Convert to writeAndFlush(...) so the promise will be notified by the transport.
+                ctx.writeAndFlush(msg, promise);
+            }
+        }, new ChannelOutboundHandlerAdapter());
+
+        final PendingWriteQueue queue = new PendingWriteQueue(channel.pipeline().lastContext());
+
+        ChannelPromise promise = channel.newPromise();
+        queue.add(1L, promise);
+        queue.add(2L, channel.voidPromise());
+        queue.removeAndWriteAll();
+
+        assertTrue(channel.finish());
+        assertTrue(promise.isDone());
+        assertTrue(promise.isSuccess());
+        assertEquals(1L, channel.readOutbound());
+        assertEquals(2L, channel.readOutbound());
+    }
+
     @Test
     public void testRemoveAndFailAllReentrantWrite() {
         final List<Integer> failOrder = Collections.synchronizedList(new ArrayList<Integer>());
@@ -377,5 +401,7 @@ public class PendingWriteQueueTest {
         }
     }
 
-    private static final class TestException extends Exception { }
+    private static final class TestException extends Exception {
+        private static final long serialVersionUID = -9018570103039458401L;
+    }
 }
