@@ -54,6 +54,17 @@ import io.netty.util.concurrent.ScheduledFuture;
 
 public class EmbeddedChannelTest {
 
+    @Test
+    public void testParent() {
+        EmbeddedChannel parent = new EmbeddedChannel();
+        EmbeddedChannel channel = new EmbeddedChannel(parent, EmbeddedChannelId.INSTANCE, true, false);
+        assertSame(parent, channel.parent());
+        assertNull(parent.parent());
+
+        assertFalse(channel.finish());
+        assertFalse(parent.finish());
+    }
+
     @Test
     public void testNotRegistered() throws Exception {
         EmbeddedChannel channel = new EmbeddedChannel(false, false);
@@ -281,6 +292,17 @@ public class EmbeddedChannelTest {
         assertNull(handler.pollEvent());
     }
 
+    @Test
+    public void testHasNoDisconnectSkipDisconnect() throws InterruptedException {
+        EmbeddedChannel channel = new EmbeddedChannel(false, new ChannelOutboundHandlerAdapter() {
+            @Override
+            public void close(ChannelHandlerContext ctx, ChannelPromise promise) throws Exception {
+                promise.tryFailure(new Throwable());
+            }
+        });
+        assertFalse(channel.disconnect().isSuccess());
+    }
+
     @Test
     public void testFinishAndReleaseAll() {
         ByteBuf in = Unpooled.buffer();
